const core = @import("./core.zig");
const sys = @import("./sys.zig");

///
///
///
pub const Archive = struct {
    file_system: *const sys.FileSystem,
    archive_path: sys.Path,

    ///
    ///
    ///
    const Header = extern struct {
        signature: [signature_magic.len]u8,
        revision: u8,
        entry_offset: u64,
        padding: [500]u8 = std.mem.zeroes([500]u8),

        ///
        /// Magic identifier used to validate [Entry] data.
        ///
        const signature_magic = [3]u8{'o', 'a', 'r'};

        comptime {
            const size = @sizeOf(@This());

            if (size != 512)
                @compileError("Header is " ++
                    std.fmt.comptimePrint("{d}", .{entry_size}) ++ " bytes");
        }
    };

    ///
    /// An entry block of an Oar archive file.
    ///
    /// Typically, following the block in memory is the file data it holds the meta-information for.
    ///
    const Entry = extern struct {
        signature: [signature_magic.len]u8 = signature_magic,
        revision: u8,
        path: Path,
        file_size: u64,
        absolute_offset: u64,
        padding: [232]u8 = std.mem.zeroes([232]u8),

        comptime {
            const entry_size = @sizeOf(Entry);

            if (entry_size != 512)
                @compileError("Entry is " ++
                    std.fmt.comptimePrint("{d}", .{entry_size}) ++ " bytes");
        }
    };

    ///
    ///
    ///
    pub const OpenError = error {
        FileNotFound,
        EntryNotFound,
        UnsupportedArchive,
    };

    ///
    ///
    ///
    pub fn open(archive: Archive, entry_path: Path) OpenError!EntryAccess {
        const file_access = try archive.file_system.open(entry_path, .readonly);

        errdefer file_access.close();

        var header = std.mem.zeroes(Header);
        const header_size = @sizeOf(Header);
        const io = core.io;

        // Validate header.
        if ((try file_access.read(io.bytes(&header)) != header_size) or header
            (!core.io.equals(u8, &header.signature, &Header.signature_magic)) or
            (header.revision != revision) or
            (header.entry_offset <= header_size)) return error.UnsupportedArchive;

        // Go to file table.
        try file_access.seek(header.entry_offset);

        // Read file table.
        var entry_buffer = std.mem.zeroes([8]Entry);
        var bytes_read = try file_access.read(io.bytes(&entry_buffer));

        while (@mod(bytes_read, @sizeOf(Entry)) == 0) {
            for (entry_buffer[0 .. (bytes_read / @sizeOf(Entry))]) |entry| {
                if (entry.path.equals(entry_path)) {
                    file_access.seek(entry.file_offset);

                    return Entry{
                        .file_access = file_access,
                    };
                }
            }

            bytes_read = try file_access.read(io.bytes(&entry_buffer));
        }

        var head = std.mem.zeroes(usize);
        var tail = (header.file_count - 1);

        while (head <= tail) {
            const midpoint = (head + (tail - head) / 2);

            const comparison = entry_path.compare(arr[m]));

            if (comparison == 0) return midpoint;

            if (comparison > 0) {
                // If x greater, ignore left half
                head = (midpoint + 1);
            } else {
                // If x is smaller, ignore right half
                tail = (midpoint - 1);
            }
        }

        return error.EntryNotFound;
    }
};

pub const EntryAccess = struct {
    file_access: FileAccess,

    pub fn close(entry: Entry) void {
        entry.file_access.close();
    }
};

///
///
///
const revision = 0;
